"""Admin for django_fido authenticator."""
from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.core.exceptions import ValidationError
from django.db import models
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.urls import reverse, reverse_lazy

from django_fido.forms import Fido2RegistrationForm
from django_fido.models import Authenticator
from django_fido.views import Fido2RegistrationRequestView, Fido2RegistrationView


class Fido2RegistrationAdminForm(Fido2RegistrationForm):
    """Registration form with user selection."""

    user = forms.ModelChoiceField(queryset=get_user_model().objects.order_by(get_user_model().USERNAME_FIELD))

    field_order = ('user', 'label')


class Fido2RegistrationRequestAdminView(PermissionRequiredMixin, Fido2RegistrationRequestView):
    """Registration request view with user selection."""

    permission_required = 'django_fido.add_authenticator'

    def get_user(self):
        """Get user based on POST request."""
        user_id = get_user_model()._meta.pk.to_python(self.request.GET.get('user'))
        try:
            return get_user_model().objects.get(pk=user_id)
        except get_user_model().DoesNotExist:
            raise Http404('User does not exist')


class AuthenticatorAddView(PermissionRequiredMixin, Fido2RegistrationView):
    """Authenticator add view."""

    permission_required = 'django_fido.add_authenticator'

    form_class = Fido2RegistrationAdminForm
    template_name = 'django_fido/add_authenticator.html'
    fido2_request_url = reverse_lazy('admin:django_fido_registration_request')
    extra_context = None

    def form_valid(self, form: forms.Form) -> HttpResponse:
        """Complete the FIDO token registration."""
        try:
            # Return value is ignored, because we need whole attestation.
            self.complete_registration(form)
        except ValidationError as error:
            form.add_error(None, error)
            return self.form_invalid(form)

        authenticator = Authenticator.objects.create(
            user=form.cleaned_data['user'],
            attestation=form.cleaned_data['attestation'],
            label=form.cleaned_data.get('label'),
        )
        return HttpResponseRedirect(reverse('admin:django_fido_authenticator_change', args=(authenticator.pk,)))

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.extra_context:
            context.update(self.extra_context)
        return context


class AuthenticatorAdmin(admin.ModelAdmin):
    """Authenticator admin."""

    list_display = ('label', 'user', 'create_datetime')
    readonly_fields = ('user', 'credential_id_data', 'attestation_data', 'counter')
    formfield_overrides = {
        models.TextField: {'widget': forms.TextInput},
    }

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            url(
                '^registration/request/$',
                self.admin_site.admin_view(Fido2RegistrationRequestAdminView.as_view()),
                name='django_fido_registration_request',
            )
        ]
        return my_urls + urls

    def add_view(self, request, form_url='', extra_context=None):
        """Customize authenticator add view."""
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta,
            'add': True,
            'has_view_permission': True,
        }
        context.update(extra_context or {})
        return AuthenticatorAddView.as_view(extra_context=context)(request)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-01 14:38
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0010_credential_id_not_null'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='authenticator',
            name='credential_data',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.27 on 2020-01-27 09:27
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0011_remove_credential_data'),
    ]

    operations = [
        migrations.AddField(
            model_name='authenticator',
            name='label',
            field=models.TextField(blank=True, max_length=255, null=True),
        ),
    ]



"""Models for storing keys."""
import base64
import json
import warnings
from binascii import b2a_hex
from datetime import date
from operator import methodcaller
from typing import List, Optional, cast
from uuid import UUID

from cryptography.hazmat.backends import default_backend
from cryptography.x509 import ExtensionNotFound, SubjectKeyIdentifier, load_der_x509_certificate
from cryptography.x509.oid import ExtensionOID
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django.db import models
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from fido2.ctap2 import AttestationObject, AttestedCredentialData
from OpenSSL import crypto

from django_fido.constants import NULL_AAGUID, PEM_CERT_TEMPLATE, AuthLevel, AuthVulnerability

# Deprecated, kept for migrations
# https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-javascript-api-v1.2-ps-20170411.html#u2f-transports
TRANSPORT_CHOICES = (
    ('bt', _('Bluetooth Classic (Bluetooth BR/EDR)')),
    ('ble', _('Bluetooth Low Energy (Bluetooth Smart)')),
    ('nfc', _('Near-Field Communications')),
    ('usb', _('USB HID')),
    ('usb-internal', _('Non-removable USB HID')),
)


@deconstructible
class TransportsValidator(object):
    """Validator for comma separated transport values.

    @ivar choices: List/tuple of available values.
    """

    choices = tuple(choice for choice, label in TRANSPORT_CHOICES)
    code = 'invalid_choice'
    message = _('Select a valid choice. %(value)s is not one of the available choices.')

    def __init__(self, choices=None, code=None, message=None):
        """Set custom `choices`, `code` or `message`."""
        warnings.warn("TransportsValidator is deprecated. It is kept only for migrations.", DeprecationWarning)
        if choices is not None:
            self.choices = choices
        if code is not None:
            self.code = code
        if message is not None:
            self.message = message

    def __call__(self, value):
        """Validate the input."""
        for chunk in force_text(value).split(','):
            if chunk not in self.choices:
                raise ValidationError(self.message, code=self.code, params={'value': chunk})


class Authenticator(models.Model):
    """Represents a registered FIDO2 authenticator.

    Autheticator fields, see https://www.w3.org/TR/webauthn/#sec-authenticator-data
     * credential_id_data - base64 encoded credential ID https://www.w3.org/TR/webauthn/#credential-id
       * This field should be used for readonly purposes only.
     * attestation_data - base64 encoded attestation object
     * counter
    """

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='authenticators', on_delete=models.CASCADE)
    create_datetime = models.DateTimeField(auto_now_add=True)

    credential_id_data = models.TextField(unique=True)
    attestation_data = models.TextField()
    counter = models.BigIntegerField(default=0)
    label = models.TextField(max_length=255, blank=True)

    class Meta:
        unique_together = [['user', 'label']]

    @property
    def credential_id(self) -> bytes:
        """Return raw credential ID."""
        return base64.b64decode(self.credential_id_data)

    @property
    def credential(self) -> AttestedCredentialData:
        """Return AttestedCredentialData object."""
        return self.attestation.auth_data.credential_data

    @property
    def attestation(self) -> AttestationObject:
        """Return AttestationObject object."""
        return AttestationObject(base64.b64decode(self.attestation_data))

    @attestation.setter
    def attestation(self, value: AttestationObject):
        self.attestation_data = base64.b64encode(value).decode('utf-8')
        self.credential_id_data = base64.b64encode(value.auth_data.credential_data.credential_id).decode('utf-8')

    def _get_metadata(self) -> Optional['AuthenticatorMetadata']:
        """Get the appropriate metadata."""
        # First test the presence of aaguid - FIDO 2
        if hasattr(self.attestation.auth_data, 'credential_data') and \
                self.attestation.auth_data.credential_data is not None and \
                self.attestation.auth_data.credential_data.aaguid != NULL_AAGUID:
            identifier = str(UUID(b2a_hex(self.credential.aaguid).decode()))
            try:
                return AuthenticatorMetadata.objects.get(identifier=identifier)
            except AuthenticatorMetadata.DoesNotExist:
                return None
        else:
            # FIXME: Add handling for UAF devices with AAID
            # Get the certificate FIDO U2F
            if 'x5c' in self.attestation.att_statement:
                cert = self.attestation.att_statement['x5c'][0]
                certificate = load_der_x509_certificate(cert, default_backend())
            else:
                # ECDSAA attestation or self attestation?
                return None
            try:
                extension = certificate.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_KEY_IDENTIFIER)
                subject_identifier = cast(SubjectKeyIdentifier, extension.value)
            except ExtensionNotFound:
                subject_identifier = SubjectKeyIdentifier.from_public_key(certificate.public_key())
            identifier = b2a_hex(subject_identifier.digest).decode()
            # Key identifiers are stored as lists...
            try:
                return AuthenticatorMetadata.objects.get(identifier__contains=identifier)
            except AuthenticatorMetadata.DoesNotExist:
                return None

    def _prepare_store(self, root_certs: List[str]) -> crypto.X509Store:
        """Prepare crypto store for verification."""
        store = crypto.X509Store()
        for root_cert in root_certs:
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, PEM_CERT_TEMPLATE.format(root_cert).encode())
            store.add_cert(cert)
        if len(self.attestation.att_statement['x5c']) > 1:
            for interm_cert in self.attestation.att_statement['x5c'][1:]:
                try:
                    store.add_cert(crypto.load_certificate(crypto.FILETYPE_ASN1, interm_cert))
                except crypto.Error:
                    # The certificate failed to load, ignore as if it is a missing link, the verification will fail
                    # It is possible that it is defined here as well as in metadata
                    pass
        return store

    @cached_property
    def metadata(self) -> Optional['AuthenticatorMetadata']:
        """Verify and return the appropriate metada for this authenticator."""
        try:
            metadata = self._get_metadata()
        except MultipleObjectsReturned:
            metadata = None
        if metadata is None or 'x5c' not in self.attestation.att_statement:
            return metadata
        # Take the device certificate and try to validate against all certs in MDS
        device_cert = crypto.load_certificate(crypto.FILETYPE_ASN1, self.attestation.att_statement['x5c'][0])
        if metadata.detailed_metadata_entry != '':
            root_certs = json.loads(metadata.detailed_metadata_entry)['attestationRootCertificates']
        else:
            root_certs = json.loads(metadata.metadata_entry).get('metadataStatement',
                                                                 {}).get('attestationRootCertificates', [])
            if not root_certs:
                return metadata
        store = self._prepare_store(root_certs)
        store_ctx = crypto.X509StoreContext(store, device_cert)
        try:
            store_ctx.verify_certificate()
        except crypto.X509StoreContextError:
            # The device certificate cannot be verified using the MDS certificate, do not trust the metadata
            return None
        return metadata


class AuthenticatorMetadata(models.Model):
    """Stores information from metadata service."""

    url = models.URLField(unique=True, null=True, blank=True)
    identifier = models.TextField(unique=True)
    metadata_entry = models.TextField()
    detailed_metadata_entry = models.TextField()

    @cached_property
    def level(self) -> AuthLevel:
        """Return last valid certification level."""
        decoded = json.loads(self.metadata_entry)
        status_dict = sorted(decoded['statusReports'],
                             key=methodcaller('get', 'effectiveDate', date.today().isoformat()),
                             reverse=True)
        # The last status should be valid
        for status in status_dict:
            # Is it directly a level?
            if status['status'] in tuple(AuthLevel):
                return AuthLevel(status['status'])
            elif status['status'] == 'REVOKED':
                return AuthLevel.NONE
        return AuthLevel.NONE

    @cached_property
    def vulnerabilities(self) -> List[AuthVulnerability]:
        """Return a list of reported vulnerabilities."""
        decoded = json.loads(self.metadata_entry)
        vulnerabilities = tuple(AuthVulnerability)
        return [AuthVulnerability(s['status']) for s in reversed(decoded['statusReports'])
                if s['status'] in vulnerabilities]

    @cached_property
    def is_update_available(self) -> bool:
        """Return whether an update is available."""
        decoded = json.loads(self.metadata_entry)
        return 'UPDATE_AVAILABLE' in [status['status'] for status in decoded['statusReports']]



"""
Django FIDO constants.

Session key to store FIDO2 requests: FIDO2_REQUEST_SESSION_KEY

Session key to store user PK for django fido authentication
 * AUTHENTICATION_USER_SESSION_KEY

FIDO 2 request type identifiers:
 * FIDO2_REGISTRATION_REQUEST
 * FIDO2_AUTHENTICATION_REQUEST
These are shared between code and JS.
"""
from enum import Enum, unique

FIDO2_REQUEST_SESSION_KEY = 'fido2_request'

AUTHENTICATION_USER_SESSION_KEY = 'django_fido_user'

FIDO2_REGISTRATION_REQUEST = 'registration'
FIDO2_AUTHENTICATION_REQUEST = 'authentication'

PEM_CERT_TEMPLATE = '-----BEGIN CERTIFICATE-----\n{}\n-----END CERTIFICATE-----\n'
NULL_AAGUID = b'\x00' * 16

# According to RFC 7518
HASH_ALG_MAPPING = {
    'HS256': 'SHA256',
    'HS384': 'SHA384',
    'HS512': 'SHA512',
    'ES256': 'SHA256',
    'ES384': 'SHA384',
    'ES512': 'SHA512',
    'RS256': 'SHA256',
    'RS384': 'SHA384',
    'RS512': 'SHA512',
}


@unique
class AuthLevel(str, Enum):
    """Enum for Levels of certification."""

    NONE = 'NOT_FIDO_CERTIFIED'
    SELF = 'SELF_ASSERTION_SUBMITTED'
    L0 = 'FIDO_CERTIFIED'
    L1 = 'FIDO_CERTIFIED_L1'
    L2 = 'FIDO_CERTIFIED_L2'
    L3 = 'FIDO_CERTIFIED_L3'
    L4 = 'FIDO_CERTIFIED_L4'
    L5 = 'FIDO_CERTIFIED_L5'


@unique
class AuthVulnerability(str, Enum):
    """Enum for Authenticator vulnerabilities."""

    REVOKED = 'REVOKED'
    USER_BYPASS = 'USER_VERIFICATION_BYPASS'
    ATTESTATION_COMPROMISE = 'ATTESTATION_KEY_COMPROMISE'
    REMOTE_COMPROMISE = 'USER_KEY_REMOTE_COMPROMISE'
    PHYSICAL_COMPROMISE = 'USER_KEY_PHYSICAL_COMPROMISE'



from unittest import skipIf
from unittest.mock import patch

from django import VERSION as DJANGO_VERSION
from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse, reverse_lazy

import django_fido.admin.authenticator
from django_fido.models import Authenticator

from .data import ATTESTATION_OBJECT, REGISTRATION_CLIENT_DATA, USER_FIRST_NAME, USER_LAST_NAME, USERNAME

User = get_user_model()


@override_settings(ROOT_URLCONF='django_fido.tests.urls')
class TestFido2RegistrationRequestAdminView(TestCase):

    url = reverse_lazy('admin:django_fido_registration_request')

    def setUp(self):
        self.user = User.objects.create_user(
            USERNAME, first_name=USER_FIRST_NAME, last_name=USER_LAST_NAME,
            is_staff=True,
        )
        self.superuser = User.objects.create_superuser(
            'admin',
            first_name='Artem', last_name='Aardvark',
            email='admin@example.com', password='passwd',
        )

    def test_anonymous(self):
        response = self.client.get(self.url)
        self.assertRedirects(response, '/admin/login/?next={}'.format(self.url), fetch_redirect_response=False)

    @skipIf(DJANGO_VERSION < (2, 1), "Skip for old Django versions.")
    def test_permission_denied(self):
        self.client.force_login(self.user)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 403)

    def test_get(self):
        self.client.force_login(self.superuser)

        response = self.client.get(self.url, data={'user': str(self.user.pk)})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()['publicKey']['user'], {
            'displayName': 'Kryten 2X4B-523P',
            'id': 'kryten',
            'name': 'kryten',
        })

    def test_get_user_not_exist(self):
        self.client.force_login(self.superuser)
        response = self.client.get(self.url, data={'user': '42'})
        self.assertEqual(response.status_code, 404)


@override_settings(ROOT_URLCONF='django_fido.tests.urls')
class TestAuthenticatorAddView(TestCase):

    url = reverse_lazy('admin:django_fido_authenticator_add')

    def setUp(self):
        self.user = User.objects.create_user(
            USERNAME, first_name=USER_FIRST_NAME, last_name=USER_LAST_NAME,
            is_staff=True,
        )
        self.superuser = User.objects.create_superuser(
            'admin',
            first_name='Artem', last_name='Aardvark',
            email='admin@example.com', password='passwd',
        )

    @skipIf(DJANGO_VERSION < (2, 1), "Skip for old Django versions.")
    def test_permission_denied(self):
        self.client.force_login(self.user)

        response_get = self.client.get(self.url)
        response_post = self.client.post(self.url)

        self.assertEqual(response_get.status_code, 403)
        self.assertEqual(response_post.status_code, 403)

    @patch.object(django_fido.admin.authenticator.AuthenticatorAddView, 'complete_registration', return_value=None)
    def test_post(self, mock):
        self.client.force_login(self.superuser)

        response = self.client.post(self.url, data={
            'user': self.user.pk,
            'client_data': REGISTRATION_CLIENT_DATA,
            'attestation': ATTESTATION_OBJECT,
            'label': 'My key',
        })

        self.assertRedirects(response, reverse('admin:django_fido_authenticator_change', args=(1,)))
        self.assertQuerysetEqual(
            Authenticator.objects.values_list('pk', 'user__username', 'attestation_data', 'label'),
            [(1, 'kryten', ATTESTATION_OBJECT, 'My key')],
            transform=tuple,
        )

    def test_get(self):
        self.client.force_login(self.superuser)
        response = self.client.get(self.url)
        choices = list(response.context['form'].fields['user'].choices)
        self.assertEqual([choice[1] for choice in choices][1:], sorted(['admin', USERNAME]))



"""Authentication backends for Django FIDO."""
import base64
import logging
from typing import Any, Dict, Optional

from django.contrib import messages
from django.contrib.auth import get_backends, get_user_model
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.base_user import AbstractBaseUser
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest
from django.utils.translation import gettext_lazy as _
from fido2.server import Fido2Server

from .settings import SETTINGS

_LOGGER = logging.getLogger(__name__)


def is_fido_backend_used() -> bool:
    """Detect whether FIDO2 authentication backend is used."""
    for auth_backend in get_backends():
        if isinstance(auth_backend, (Fido2AuthenticationBackend, Fido2GeneralAuthenticationBackend)):
            return True

    return False


class Fido2AuthenticationBackend(object):
    """
    Authenticate user using FIDO 2.

    @cvar counter_error_message: Error message in case FIDO 2 device counter didn't increase.
    """

    counter_error_message = _("Counter of the FIDO 2 device decreased. Device may have been duplicated.")

    def authenticate(self, request: HttpRequest, user: AbstractBaseUser, fido2_server: Fido2Server,
                     fido2_state: Dict[str, bytes], fido2_response: Dict[str, Any]) -> Optional[AbstractBaseUser]:
        """Authenticate using FIDO 2."""
        credentials = [a.credential for a in user.authenticators.all()]
        try:
            credential = fido2_server.authenticate_complete(
                fido2_state, credentials, fido2_response['credential_id'], fido2_response['client_data'],
                fido2_response['authenticator_data'], fido2_response['signature'])
        except ValueError as error:
            _LOGGER.info("FIDO 2 authentication failed with error: %r", error)
            return None

        device = user.authenticators.get(credential_id_data=base64.b64encode(credential.credential_id).decode('utf-8'))
        try:
            self.mark_device_used(device, fido2_response['authenticator_data'].counter)
        except ValueError:
            # Raise `PermissionDenied` to stop the authentication process and skip remaining backends.
            messages.error(request, self.counter_error_message)
            raise PermissionDenied("Counter didn't increase.")
        return user

    def mark_device_used(self, device, counter):
        """Update FIDO 2 device usage information."""
        if counter == 0 and device.counter == 0:
            # Counter is unsupported by the device, bail out early
            return
        if counter <= device.counter:
            _LOGGER.info("FIDO 2 authentication failed because of not increasing counter.")
            raise ValueError("Counter didn't increase.")
        device.counter = counter
        device.full_clean()
        device.save()

    def get_user(self, user_id):
        """Return user based on its ID."""
        try:
            return get_user_model().objects.get(pk=user_id)
        except get_user_model().DoesNotExist:
            return None


class Fido2GeneralAuthenticationBackend(ModelBackend):
    """Authenticated user using any username-password backend and Fido2AuthenticationBackend."""

    def __init__(self, *args, **kwargs):
        """Initialize backend."""
        super().__init__(*args, **kwargs)
        self.fido_backend = Fido2AuthenticationBackend()

    def authenticate(
        self,
        request: HttpRequest,
        username: str,
        password: str,
        fido2_server: Fido2Server,
        fido2_state: Dict[str, bytes],
        fido2_response: Dict[str, Any],
        **kwargs
    ) -> Optional[AbstractBaseUser]:
        """Authenticate using username, password and FIDO 2 token."""
        for auth_backend in SETTINGS.authentication_backends:
            user = auth_backend().authenticate(request, username=username, password=password, **kwargs)
            if user is not None:
                return self.fido_backend.authenticate(request, user, fido2_server, fido2_state, fido2_response)

        return None



"""Command to download metadata for authenticators."""
import base64
import hashlib
import json
from typing import Any, Dict, List, Tuple

import requests
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from django.core.management.base import BaseCommand, CommandError
from django.db.transaction import atomic
from jwcrypto.jwk import JWK
from jwcrypto.jws import InvalidJWSSignature
from jwcrypto.jwt import JWT
from OpenSSL import crypto

from django_fido.constants import HASH_ALG_MAPPING, PEM_CERT_TEMPLATE
from django_fido.models import AuthenticatorMetadata
from django_fido.settings import SETTINGS


class InvalidCert(Exception):
    """Raised when certificate validation fails."""


def urlsafe_b64decode(decodable: bytes) -> bytes:
    """Alternative implementation to fill the necessary padding."""
    m = len(decodable) % 4
    if m == 2:
        decodable += b'=='
    elif m == 3:
        decodable += b'='
    return base64.urlsafe_b64decode(decodable)


def _prepare_crypto_store(jwt: JWT) -> crypto.X509Store:
    """Prepare crytpographic store for verification."""
    # Create crypto context
    store = crypto.X509Store()
    for key in jwt.token.jose_header['x5c'][1:]:
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, PEM_CERT_TEMPLATE.format(key).encode())
        store.add_cert(cert)
    for root_cert_file in SETTINGS.metadata_service['certificate']:
        with open(str(root_cert_file), 'rb') as root_file:
            root_cert = crypto.load_certificate(crypto.FILETYPE_PEM, root_file.read())
    store.add_cert(root_cert)
    # CRL handling
    # FIXME: This part is not tested in unittests (intentionally) as it might be more suited for integration testing
    for crl_file in SETTINGS.metadata_service['crl_list']:
        with open(str(crl_file), 'rb') as file:
            crl_list = x509.load_pem_x509_crl(file.read(), default_backend())
        store.add_crl(crypto.CRL.from_cryptography(crl_list))
    if SETTINGS.metadata_service['crl_list']:
        store.set_flags(crypto.X509StoreFlags.CRL_CHECK)
    return store


def verify_certificate(jwt: JWT) -> JWK:
    """Get (and verify) the signing key from JWT."""
    # First element in the header is our actual key
    try:
        decoding_key = JWK.from_pem(PEM_CERT_TEMPLATE.format(jwt.token.jose_header['x5c'][0]).encode())
    except ValueError:
        raise InvalidCert('Cannot decode key.')
    if SETTINGS.metadata_service['disable_cert_verification']:
        return decoding_key
    if not SETTINGS.metadata_service['certificate']:
        raise CommandError("Certificate verification enabled, but no certificate set. "
                           "Please set certificate or disable validation.")
    # Create context and verify
    store = _prepare_crypto_store(jwt)
    decoding_cert = crypto.load_certificate(crypto.FILETYPE_PEM,
                                            PEM_CERT_TEMPLATE.format(jwt.token.jose_header['x5c'][0]).encode())
    store_ctx = crypto.X509StoreContext(store, decoding_cert)
    try:
        store_ctx.verify_certificate()
    except crypto.X509StoreContextError:
        raise InvalidCert('Key could not be verified.')
    else:
        return decoding_key


def _get_metadata() -> Tuple[Dict[str, Any], str]:
    """Download the metadata TOC."""
    try:
        metadata = requests.get(SETTINGS.metadata_service['url'],
                                params={'token': SETTINGS.metadata_service['access_token']},
                                timeout=SETTINGS.metadata_service['timeout'])
    except requests.exceptions.RequestException:
        raise CommandError('MDS response error.')
    # First, we decode the unverified headers to get the certificate
    try:
        decoded_jwt = JWT(jwt=metadata.content.decode())
    except ValueError:
        raise CommandError('MDS response malformed.')
    # x5c element in header contains the signing certificate and possibly intermediate certificates
    # Use the first one to verify signature, the others can be used to verify the first one
    try:
        decoding_key = verify_certificate(decoded_jwt)
    except InvalidCert:
        raise CommandError('Could not read the key.')
    try:
        decoded_jwt.deserialize(metadata.content.decode(), key=decoding_key)
    except InvalidJWSSignature:
        raise CommandError('Could not verify MDS signature.')
    # Return parsed metadata and the algorith for signing
    return json.loads(decoded_jwt.claims), json.loads(decoded_jwt.header)['alg']


class Command(BaseCommand):
    """Download metadata for authenticators."""

    help = "Download metadata for authenticators."""

    def add_arguments(self, parser):
        """Parse command arguments."""

    def _update_auth(self, authenticator_data: Dict[str, Any], identifiers: List[str], hash_alg: str) -> List[str]:
        """Update individual authenticators from MDS."""
        downloaded_identifiers = []
        mds3 = SETTINGS.metadata_service['mds_format'] == 3
        if not mds3:
            url = authenticator_data['url']
            # Cleanup the old instances
            authenticator = AuthenticatorMetadata.objects.filter(url=url).delete()
            auth_metadata = requests.get(url, params={'token': SETTINGS.metadata_service['access_token']},
                                         timeout=SETTINGS.metadata_service['timeout'])
            hash = hashlib.new(hash_alg, auth_metadata.content)
            if hash.digest() != urlsafe_b64decode(authenticator_data['hash'].encode()):
                self.stderr.write('Hash invalid for authenticator {}.'.format(identifiers))
                detailed_metadata_entry = ''
            else:
                detailed_metadata_entry = urlsafe_b64decode(auth_metadata.content).decode()
        for ident in identifiers:
            downloaded_identifiers += [ident]
            authenticator, _ = AuthenticatorMetadata.objects.get_or_create(identifier=ident)
            authenticator.metadata_entry = json.dumps(authenticator_data)
            if not mds3:
                authenticator.detailed_metadata_entry = detailed_metadata_entry
            authenticator.save()
        return downloaded_identifiers

    def handle(self, **options):
        """Download and parse metadata from metadata service."""
        if SETTINGS.metadata_service is None or (
                SETTINGS.metadata_service['access_token'] is None and SETTINGS.metadata_service['mds_format'] == 2):
            raise CommandError('Access token must be specified for MDS2.')

        downloaded_identifiers = []
        metadata, alg = _get_metadata()
        # Convert alg name to corresponding hashing algorithm
        hash_alg = HASH_ALG_MAPPING.get(alg)
        if hash_alg is None or hash_alg.lower() not in hashlib.algorithms_available:
            raise CommandError('Unsupported hash algorithm {}.'.format(alg))
        for authenticator_data in metadata['entries']:
            if 'aaid' in authenticator_data:
                identifier = [authenticator_data['aaid']]
                downloaded_identifiers += identifier
            elif 'aaguid' in authenticator_data:
                identifier = [authenticator_data['aaguid']]
                downloaded_identifiers += identifier
            elif 'attestationCertificateKeyIdentifiers' in authenticator_data:
                identifier = authenticator_data['attestationCertificateKeyIdentifiers']
                downloaded_identifiers += identifier
            else:
                self.stderr.write('Cannot determine the identificator from metadata response.')
                continue
            with atomic():
                downloaded_identifiers += self._update_auth(authenticator_data, identifier, hash_alg)
        # Cleanup authenticators that were not present in the donwload
        AuthenticatorMetadata.objects.exclude(identifier__in=downloaded_identifiers).delete()



"""URLs for django_fido application tests."""
from django.conf.urls import include, url
from django.contrib import admin

from django_fido.views import AttestationConveyancePreference, Fido2RegistrationView

urlpatterns = [
    url('', include('django_fido.urls')),
    url('admin/', admin.site.urls),
    url('direct_registration/',
        Fido2RegistrationView.as_view(attestation=AttestationConveyancePreference.DIRECT),
        name='registration_direct'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-05-11 08:57
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_fido', '0013_unique_user_label'),
    ]

    operations = [
        migrations.CreateModel(
            name='AuthenticatorMetadata',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.URLField()),
                ('identifier', models.TextField(unique=True)),
                ('metadata_entry', models.TextField()),
                ('detailed_metadata_entry', models.TextField()),
            ],
        ),
    ]



"""Unittests for download_authenticator_metadata."""
import json
import os
from base64 import urlsafe_b64encode
from datetime import datetime
from io import StringIO
from unittest.mock import patch

import responses
from django.conf import settings
from django.core.management import CommandError, call_command
from django.test import SimpleTestCase, TestCase, override_settings
from jwcrypto.jwk import JWK
from jwcrypto.jwt import JWT
from requests.exceptions import RequestException

from django_fido.management.commands.download_authenticator_metadata import (InvalidCert, _get_metadata,
                                                                             _prepare_crypto_store, verify_certificate)
from django_fido.models import AuthenticatorMetadata
from django_fido.settings import SETTINGS

DIR_PATH = os.path.join(os.path.dirname(__file__), 'data', 'mds')


def get_file_content(path):
    """Return file content."""
    with open(path) as f:
        content = f.read()
    return content.strip().encode()


def prepare_timed_context(jwt):
    """Return crypto context frozen in time."""
    store = _prepare_crypto_store(jwt)
    store.set_time(datetime(2020, 12, 12, 12, 0, 0))
    return store


@patch('django_fido.management.commands.download_authenticator_metadata._prepare_crypto_store', prepare_timed_context)
class TestVerifyCertificate(SimpleTestCase):
    """Unittests for verify_certificate.

    These tests use a real world certificate that expires in 2045.
    This seems quite far in the future for now, but beware!
    """

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token', 'DISABLE_CERT_VERIFICATION': True})
    def test_disabled_verification(self):
        jwt = JWT(jwt=get_file_content(os.path.join(DIR_PATH, 'correct.txt')).decode())
        key = verify_certificate(jwt)
        self.assertIsInstance(key, JWK)

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                     'DISABLE_CERT_VERIFICATION': False})
    def test_no_certificate(self):
        jwt = JWT(jwt=get_file_content(os.path.join(DIR_PATH, 'correct.txt')).decode())
        with self.assertRaises(CommandError):
            verify_certificate(jwt)

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                     'CERTIFICATE': [os.path.join(DIR_PATH, 'Root.cer')]})
    def test_correct_root_cert(self):
        jwt = JWT(jwt=get_file_content(os.path.join(DIR_PATH, 'correct.txt')).decode())
        key = verify_certificate(jwt)
        self.assertIsInstance(key, JWK)

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                     'CERTIFICATE': [os.path.join(DIR_PATH, 'Root_bad.cer')]})
    def test_bad_root_cert(self):
        jwt = JWT(jwt=get_file_content(os.path.join(DIR_PATH, 'correct.txt')).decode())
        with self.assertRaises(InvalidCert):
            verify_certificate(jwt)

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                     'CERTIFICATE': [os.path.join(DIR_PATH, 'Root_bad.cer'),
                                                                     os.path.join(DIR_PATH, 'Root.cer')]})
    def test_multiple_root_cert(self):
        jwt = JWT(jwt=get_file_content(os.path.join(DIR_PATH, 'correct.txt')).decode())
        key = verify_certificate(jwt)
        self.assertIsInstance(key, JWK)

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                     'CERTIFICATE': [os.path.join(DIR_PATH, 'NotRoot.cer')]})
    def test_not_a_root_cert(self):
        jwt = JWT(jwt=get_file_content(os.path.join(DIR_PATH, 'correct.txt')).decode())
        with self.assertRaises(InvalidCert):
            verify_certificate(jwt)

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token'})
    def test_malformed_key(self):
        token = 'eyJhbGciOiJFUzI1NiIsInR5cCI6IkpXVCIsIng1YyI6WyJhYWFhYWFhIl19.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6Ikp' \
                'vaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.Dvmv1EMeu-pAqJdJAIXKef6M_Kx2Dn2qCLZyBF63f3RcI1ddDCXADlLmwvMjCX7u' \
                'V1R5AbLMf_rLxUlGZZnXvg'
        jwt = JWT(jwt=token)
        with self.assertRaises(InvalidCert):
            verify_certificate(jwt)


@override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                 'CERTIFICATE': [os.path.join(DIR_PATH, 'Root.cer')]})
@patch('django_fido.management.commands.download_authenticator_metadata._prepare_crypto_store', prepare_timed_context)
class TestGetMetadata(SimpleTestCase):
    """Unittests for get_metadata command."""

    def test_error_response(self):
        with responses.RequestsMock() as rsps:
            with self.assertRaisesMessage(CommandError, 'MDS response error.'):
                rsps.add(responses.GET, 'https://mds2.fidoalliance.org/', body=RequestException('Some error'))
                _get_metadata()

    def test_malformed_response(self):
        with responses.RequestsMock() as rsps:
            with self.assertRaisesMessage(CommandError, 'MDS response malformed.'):
                rsps.add(responses.GET, 'https://mds2.fidoalliance.org/', body='')
                _get_metadata()

    def test_ok_response(self):
        content = get_file_content(os.path.join(DIR_PATH, 'correct.txt'))
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://mds2.fidoalliance.org/', body=content)
            metadata, alg = _get_metadata()
        self.assertEqual(set(metadata.keys()), {'entries', 'nextUpdate', 'no', 'legalHeader'})
        self.assertEqual(alg, 'ES256')

    def test_bad_signature_response(self):
        content = get_file_content(os.path.join(DIR_PATH, 'bad.txt'))
        with responses.RequestsMock() as rsps:
            with self.assertRaisesMessage(CommandError, 'Could not verify MDS signature.'):
                rsps.add(responses.GET, 'https://mds2.fidoalliance.org/', body=content)
                _get_metadata()

    def test_default_url(self):
        content = get_file_content(os.path.join(DIR_PATH, 'correct.txt'))
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://mds2.fidoalliance.org/', body=content)
            _get_metadata()

    def test_custom_url(self):
        content = get_file_content(os.path.join(DIR_PATH, 'correct.txt'))
        with override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                             'URL': 'https://example.com',
                                                             'CERTIFICATE': [os.path.join(DIR_PATH, 'Root.cer')]}):
            with responses.RequestsMock() as rsps:
                rsps.add(responses.GET, 'https://example.com', body=content)
                _get_metadata()

    def test_bad_cert(self):
        content = get_file_content(os.path.join(DIR_PATH, 'correct.txt'))
        with override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token',
                                                             'CERTIFICATE': [os.path.join(DIR_PATH, 'Root_bad.cer')]}):
            with responses.RequestsMock() as rsps:
                rsps.add(responses.GET, 'https://mds2.fidoalliance.org/', body=content)
                with self.assertRaises(CommandError):
                    _get_metadata()


@override_settings(DJANGO_FIDO_METADATA_SERVICE={'ACCESS_TOKEN': 'secret_token'})
class TestDownloadAuthenticatorMetadata(TestCase):
    """Unittests for download_authenticator_metadata management command."""

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={})
    def test_not_configured(self):
        # Signal is not sent on manipulation with individual settings, invalidate manually
        del settings.DJANGO_FIDO_METADATA_SERVICE
        SETTINGS.invalidate_cache()
        with self.assertRaisesMessage(CommandError, 'Access token must be specified for MDS2.'):
            call_command('download_authenticator_metadata')

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_empty_list(self, get_metada_patch):
        get_metada_patch.return_value = ({'entries': []}, 'ES256')
        call_command('download_authenticator_metadata')
        self.assertFalse(AuthenticatorMetadata.objects.all().exists())

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_aaid(self, get_metada_patch):
        entry = {'aaid': '1234#5678', 'url': 'https://example.com/1234abcd',
                 'hash': 'YXIP-OAlWmYNiw2TQHhPJtdWmNRIm78aitlsxhreXJA'}
        get_metada_patch.return_value = ({'entries': [entry]}, 'ES256')
        payload = {
            "description": "FIDO Alliance Sample UAF Authenticator",
            "aaid": "1234#5678",
            "authenticatorVersion": 2,
            "upv": [
                {"major": 1, "minor": 0},
                {"major": 1, "minor": 1}
            ],
            "assertionScheme": "UAFV1TLV",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [
                [{
                    "userVerification": 2,
                    "baDesc": {
                        "FAR": 0.00002,
                        "maxRetries": 5,
                        "blockSlowdown": 30,
                        "maxReferenceDataSets": 5
                    }
                }]
            ],
            "keyProtection": 6,
            "isKeyRestricted": True,
            "matcherProtection": 2,
            "cryptoStrength": 128,
            "operatingEnv": "TEEs based on ARM TrustZone HW",
            "attachmentHint": 1,
            "isSecondFactorOnly": False,
            "tcDisplay": 5,
            "tcDisplayContentType": "image/png",
            "tcDisplayPNGCharacteristics": [{
                "width": 320,
                "height": 480,
                "bitDepth": 16,
                "colorType": 2,
                "compression": 0,
                "filter": 0,
                "interlace": 0
            }],
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
             ]
        }
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://example.com/1234abcd',
                     body=urlsafe_b64encode(json.dumps(payload, sort_keys=True).encode()))
            call_command('download_authenticator_metadata')
        metadata = AuthenticatorMetadata.objects.get(identifier='1234#5678')
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertJSONEqual(metadata.detailed_metadata_entry, payload)

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_aaid_update(self, get_metada_patch):
        AuthenticatorMetadata.objects.create(identifier='1234#5678', url='https://example.com/1234abcd')
        entry = {'aaid': '1234#5678', 'url': 'https://example.com/1234abcd',
                 'hash': 'YXIP-OAlWmYNiw2TQHhPJtdWmNRIm78aitlsxhreXJA'}
        get_metada_patch.return_value = ({'entries': [entry]}, 'ES256')
        payload = {
            "description": "FIDO Alliance Sample UAF Authenticator",
            "aaid": "1234#5678",
            "authenticatorVersion": 2,
            "upv": [
                {"major": 1, "minor": 0},
                {"major": 1, "minor": 1}
            ],
            "assertionScheme": "UAFV1TLV",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [
                [{
                    "userVerification": 2,
                    "baDesc": {
                        "FAR": 0.00002,
                        "maxRetries": 5,
                        "blockSlowdown": 30,
                        "maxReferenceDataSets": 5
                    }
                }]
            ],
            "keyProtection": 6,
            "isKeyRestricted": True,
            "matcherProtection": 2,
            "cryptoStrength": 128,
            "operatingEnv": "TEEs based on ARM TrustZone HW",
            "attachmentHint": 1,
            "isSecondFactorOnly": False,
            "tcDisplay": 5,
            "tcDisplayContentType": "image/png",
            "tcDisplayPNGCharacteristics": [{
                "width": 320,
                "height": 480,
                "bitDepth": 16,
                "colorType": 2,
                "compression": 0,
                "filter": 0,
                "interlace": 0
            }],
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
             ]
        }
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://example.com/1234abcd',
                     body=urlsafe_b64encode(json.dumps(payload, sort_keys=True).encode()))
            call_command('download_authenticator_metadata')
        metadata = AuthenticatorMetadata.objects.get(identifier='1234#5678')
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertJSONEqual(metadata.detailed_metadata_entry, payload)

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_aaguid(self, get_metada_patch):
        entry = {'attestationCertificateKeyIdentifiers': ['7c0903708b87115b0b422def3138c3c864e44573'],
                 'url': 'https://example.com/abcd', 'hash': 'fz8OLQ8NusYLf8XopH39Hibg3wCcqFVlu_vRA-IK074'}
        get_metada_patch.return_value = ({'entries': [entry]}, 'ES256')
        payload = {
            "description": "FIDO Alliance Sample U2F Authenticator",
            "attestationCertificateKeyIdentifiers": ["7c0903708b87115b0b422def3138c3c864e44573"],
            "protocolFamily": "u2f",
            "authenticatorVersion": 2,
            "upv": [{"major": 1, "minor": 0}],
            "assertionScheme": "U2FV1BIN",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [[{"userVerification": 1}]],
            "keyProtection": 10,
            "matcherProtection": 4,
            "cryptoStrength": 128,
            "operatingEnv": "Secure Element (SE)",
            "attachmentHint": 2,
            "isSecondFactorOnly": True,
            "tcDisplay": 0,
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
                ]
            }
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://example.com/abcd',
                     body=urlsafe_b64encode(json.dumps(payload, sort_keys=True).encode()))
            call_command('download_authenticator_metadata')
        metadata = AuthenticatorMetadata.objects.get(identifier__contains='7c0903708b87115b0b422def3138c3c864e44573')
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertJSONEqual(metadata.detailed_metadata_entry, payload)

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_key_identifiers(self, get_metada_patch):
        entry = {'aaguid': '7c0903708b87115b0b422def3138c3c864e44573',
                 'url': 'https://example.com/7c0903708b87115b0b422def3138c3c864e44573',
                 'hash': 'fz8OLQ8NusYLf8XopH39Hibg3wCcqFVlu_vRA-IK074'}
        get_metada_patch.return_value = ({'entries': [entry]}, 'ES256')
        payload = {
            "description": "FIDO Alliance Sample U2F Authenticator",
            "attestationCertificateKeyIdentifiers": ["7c0903708b87115b0b422def3138c3c864e44573"],
            "protocolFamily": "u2f",
            "authenticatorVersion": 2,
            "upv": [{"major": 1, "minor": 0}],
            "assertionScheme": "U2FV1BIN",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [[{"userVerification": 1}]],
            "keyProtection": 10,
            "matcherProtection": 4,
            "cryptoStrength": 128,
            "operatingEnv": "Secure Element (SE)",
            "attachmentHint": 2,
            "isSecondFactorOnly": True,
            "tcDisplay": 0,
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
            ],
        }
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://example.com/7c0903708b87115b0b422def3138c3c864e44573',
                     body=urlsafe_b64encode(json.dumps(payload, sort_keys=True).encode()))
            call_command('download_authenticator_metadata')
        metadata = AuthenticatorMetadata.objects.get(identifier='7c0903708b87115b0b422def3138c3c864e44573')
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertJSONEqual(metadata.detailed_metadata_entry, payload)

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_hash_mismatch(self, get_metada_patch):
        output = StringIO()
        entry = {'aaguid': '7c0903708b87115b0b422def3138c3c864e44573',
                 'url': 'https://example.com/7c0903708b87115b0b422def3138c3c864e44573',
                 'hash': 'XXXXXXXXXX'}
        get_metada_patch.return_value = ({'entries': [entry]}, 'ES256')
        payload = {
            "description": "FIDO Alliance Sample U2F Authenticator",
            "attestationCertificateKeyIdentifiers": ["7c0903708b87115b0b422def3138c3c864e44573"],
            "protocolFamily": "u2f",
            "authenticatorVersion": 2,
            "upv": [{"major": 1, "minor": 0}],
            "assertionScheme": "U2FV1BIN",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [[{"userVerification": 1}]],
            "keyProtection": 10,
            "matcherProtection": 4,
            "cryptoStrength": 128,
            "operatingEnv": "Secure Element (SE)",
            "attachmentHint": 2,
            "isSecondFactorOnly": True,
            "tcDisplay": 0,
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
            ],
        }
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://example.com/7c0903708b87115b0b422def3138c3c864e44573',
                     body=urlsafe_b64encode(json.dumps(payload).encode()))
            call_command('download_authenticator_metadata', stderr=output)
        metadata = AuthenticatorMetadata.objects.get(identifier='7c0903708b87115b0b422def3138c3c864e44573')
        self.assertEqual("Hash invalid for authenticator ['7c0903708b87115b0b422def3138c3c864e44573'].\n",
                         output.getvalue())
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertEqual(metadata.detailed_metadata_entry, '')

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_unknown_alg(self, get_metadata_patch):
        entry = {'aaguid': '7c0903708b87115b0b422def3138c3c864e44573',
                 'url': 'https://example.com/7c0903708b87115b0b422def3138c3c864e44573',
                 'hash': 'XXXXXXXXXX'}
        get_metadata_patch.return_value = ({'entries': [entry]}, 'none')
        with self.assertRaisesRegex(CommandError, 'Unsupported hash algorithm none.'):
            call_command('download_authenticator_metadata')

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_key_unknown(self, get_metada_patch):
        output = StringIO()
        get_metada_patch.return_value = ({'entries': [{}]}, 'ES256')
        call_command('download_authenticator_metadata', stderr=output)
        self.assertEqual('Cannot determine the identificator from metadata response.\n',
                         output.getvalue())

    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_cleanup(self, get_metadata_patch):
        AuthenticatorMetadata.objects.create(identifier='AAAA', url='https://example.com/aaaa')
        AuthenticatorMetadata.objects.create(identifier='1234#5678', url='https://example.com/1234abcd')
        entry = {'aaid': '1234#5678', 'url': 'https://example.com/1234abcd',
                 'hash': 'YXIP-OAlWmYNiw2TQHhPJtdWmNRIm78aitlsxhreXJA'}
        get_metadata_patch.return_value = ({'entries': [entry]}, 'ES256')
        payload = {
            "description": "FIDO Alliance Sample UAF Authenticator",
            "aaid": "1234#5678",
            "authenticatorVersion": 2,
            "upv": [
                {"major": 1, "minor": 0},
                {"major": 1, "minor": 1}
            ],
            "assertionScheme": "UAFV1TLV",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [
                [{
                    "userVerification": 2,
                    "baDesc": {
                        "FAR": 0.00002,
                        "maxRetries": 5,
                        "blockSlowdown": 30,
                        "maxReferenceDataSets": 5
                    }
                }]
            ],
            "keyProtection": 6,
            "isKeyRestricted": True,
            "matcherProtection": 2,
            "cryptoStrength": 128,
            "operatingEnv": "TEEs based on ARM TrustZone HW",
            "attachmentHint": 1,
            "isSecondFactorOnly": False,
            "tcDisplay": 5,
            "tcDisplayContentType": "image/png",
            "tcDisplayPNGCharacteristics": [{
                "width": 320,
                "height": 480,
                "bitDepth": 16,
                "colorType": 2,
                "compression": 0,
                "filter": 0,
                "interlace": 0
            }],
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
             ]
        }
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, 'https://example.com/1234abcd',
                     body=urlsafe_b64encode(json.dumps(payload, sort_keys=True).encode()))
            call_command('download_authenticator_metadata')
        metadata = AuthenticatorMetadata.objects.get(identifier='1234#5678')
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertJSONEqual(metadata.detailed_metadata_entry, payload)
        self.assertFalse(AuthenticatorMetadata.objects.filter(identifier='AAAA').exists())

    @override_settings(DJANGO_FIDO_METADATA_SERVICE={'MDS_FORMAT': 3})
    @patch('django_fido.management.commands.download_authenticator_metadata._get_metadata')
    def test_aaid_v3(self, get_metadata_patch):
        payload = {
            "description": "FIDO Alliance Sample UAF Authenticator",
            "aaid": "1234#5678",
            "authenticatorVersion": 2,
            "upv": [
                {"major": 1, "minor": 0},
                {"major": 1, "minor": 1}
            ],
            "assertionScheme": "UAFV1TLV",
            "authenticationAlgorithm": 1,
            "publicKeyAlgAndEncoding": 256,
            "attestationTypes": [15879],
            "userVerificationDetails": [
                [{
                    "userVerification": 2,
                    "baDesc": {
                        "FAR": 0.00002,
                        "maxRetries": 5,
                        "blockSlowdown": 30,
                        "maxReferenceDataSets": 5
                    }
                }]
            ],
            "keyProtection": 6,
            "isKeyRestricted": True,
            "matcherProtection": 2,
            "cryptoStrength": 128,
            "operatingEnv": "TEEs based on ARM TrustZone HW",
            "attachmentHint": 1,
            "isSecondFactorOnly": False,
            "tcDisplay": 5,
            "tcDisplayContentType": "image/png",
            "tcDisplayPNGCharacteristics": [{
                "width": 320,
                "height": 480,
                "bitDepth": 16,
                "colorType": 2,
                "compression": 0,
                "filter": 0,
                "interlace": 0
            }],
            "attestationRootCertificates": [
                "MIICPTCCAeOgAwIBAgIJAOuexvU3Oy2wMAoGCCqGSM49BAMCMHsxIDAeBgNVBAMM"
                "F1NhbXBsZSBBdHRlc3RhdGlvbiBSb290MRYwFAYDVQQKDA1GSURPIEFsbGlhbmNl"
                "MREwDwYDVQQLDAhVQUYgVFdHLDESMBAGA1UEBwwJUGFsbyBBbHRvMQswCQYDVQQI"
                "DAJDQTELMAkGA1UEBhMCVVMwHhcNMTQwNjE4MTMzMzMyWhcNNDExMTAzMTMzMzMy"
                "WjB7MSAwHgYDVQQDDBdTYW1wbGUgQXR0ZXN0YXRpb24gUm9vdDEWMBQGA1UECgwN"
                "RklETyBBbGxpYW5jZTERMA8GA1UECwwIVUFGIFRXRywxEjAQBgNVBAcMCVBhbG8g"
                "QWx0bzELMAkGA1UECAwCQ0ExCzAJBgNVBAYTAlVTMFkwEwYHKoZIzj0CAQYIKoZI"
                "zj0DAQcDQgAEH8hv2D0HXa59/BmpQ7RZehL/FMGzFd1QBg9vAUpOZ3ajnuQ94PR7"
                "aMzH33nUSBr8fHYDrqOBb58pxGqHJRyX/6NQME4wHQYDVR0OBBYEFPoHA3CLhxFb"
                "C0It7zE4w8hk5EJ/MB8GA1UdIwQYMBaAFPoHA3CLhxFbC0It7zE4w8hk5EJ/MAwG"
                "A1UdEwQFMAMBAf8wCgYIKoZIzj0EAwIDSAAwRQIhAJ06QSXt9ihIbEKYKIjsPkri"
                "VdLIgtfsbDSu7ErJfzr4AiBqoYCZf0+zI55aQeAHjIzA9Xm63rruAxBZ9ps9z2XN"
                "lQ=="
             ]
        }
        entry = {'aaid': '1234#5678',
                 'metadataStatement': payload,
                 'hash': 'YXIP-OAlWmYNiw2TQHhPJtdWmNRIm78aitlsxhreXJA'}
        get_metadata_patch.return_value = ({'entries': [entry]}, 'ES256')
        call_command('download_authenticator_metadata')
        metadata = AuthenticatorMetadata.objects.get(identifier='1234#5678')
        self.assertJSONEqual(metadata.metadata_entry, entry)
        self.assertEqual(metadata.detailed_metadata_entry, '')
